
import numpy as np
import tpot2
import networkx as nx
from tpot2.representations.graph_pipeline_individual import GraphIndividual
import random

from tpot2.representations.graph_pipeline_individual.individual import create_node


def estimator_graph_individual_generator(
                                                root_config_dict,
                                                inner_config_dict=None,
                                                leaf_config_dict=None,
                                                max_depth = np.inf,
                                                max_size = np.inf, 
                                                max_children = np.inf,
                                                **kwargs,
                                            ) :



        n_nodes = 0
        while True:
            if n_nodes < max_size:
                    n_nodes += 1
            
            for k in root_config_dict.keys():
                
                graph = nx.DiGraph()
                root = create_node(config_dict={k:root_config_dict[k]})
                graph.add_node(root)
                
                ind = tpot2.GraphIndividual(    inner_config_dict=inner_config_dict,  
                                                    leaf_config_dict=leaf_config_dict,
                                                    root_config_dict=root_config_dict,
                                                    initial_graph = graph,
                                                    max_depth = max_depth,
                                                    max_size = max_size, 
                                                    max_children = max_children,
                                                    **kwargs,
                                                    )
                
                
                
                if n_nodes > 0:
                    for _ in range(np.random.randint(0,min(n_nodes,3))):
                        if random.random() < 0.5:
                            ind._mutate_insert_leaf()
                        else: 
                            ind._mutate_insert_inner_node()

                for _ in range(np.random.randint(0,ind.graph.number_of_nodes())):
                     ind._mutate_add_connection_from()
                
                yield ind

            

class BaggingCompositeGraphSklearn():
    def __init__(self) -> None:
        pass

class BoostingCompositeGraphSklearn():
    def __init__(self) -> None:
        pass




import random
from scipy.stats import loguniform, logser #TODO: remove this dependency?
import numpy as np #TODO: remove this dependency and use scipy instead?




#Replicating the API found in optuna: https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html
#copy-pasted some code
def suggest_categorical(name, choices):
    return random.choice(choices)

def suggest_float(
    name: str,
    low: float,
    high: float,
    *,
    step = None,
    log = False,
    ):
    
    if log and step is not None:
        raise ValueError("The parameter `step` is not supported when `log` is true.")

    if low > high:
        raise ValueError(
            "The `low` value must be smaller than or equal to the `high` value "
            "(low={}, high={}).".format(low, high)
        )

    if log and low <= 0.0:
        raise ValueError(
            "The `low` value must be larger than 0 for a log distribution "
            "(low={}, high={}).".format(low, high)
        )

    if step is not None and step <= 0:
        raise ValueError(
            "The `step` value must be non-zero positive value, " "but step={}.".format(step)
        )

    #TODO check this produces correct output
    if log:
        value = np.random.uniform(np.log(low),np.log(high))
        return np.e**value

    else:
        if step is not None:
            return np.random.choice(np.arange(low,high,step))
        else:
            return np.random.uniform(low,high)


def suggest_discrete_uniform(name, low, high, q):
    return suggest_float(name, low, high, step=q)


def suggest_int(name, low, high, step=1, log=False):
    if low == high: #TODO check that this matches optuna's behaviour
        return low
    
    if log and step >1:
        raise ValueError("The parameter `step`>1 is not supported when `log` is true.")

    if low > high:
        raise ValueError(
            "The `low` value must be smaller than or equal to the `high` value "
            "(low={}, high={}).".format(low, high)
        )

    if log and low <= 0.0:
        raise ValueError(
            "The `low` value must be larger than 0 for a log distribution "
            "(low={}, high={}).".format(low, high)
        )

    if step is not None and step <= 0:
        raise ValueError(
            "The `step` value must be non-zero positive value, " "but step={}.".format(step)
        )

    if log:
        value = np.random.uniform(np.log(low),np.log(high))
        return int(np.e**value)
    else:
        return np.random.choice(list(range(low,high,step)))

def suggest_uniform(name, low, high):
    return suggest_float(name, low, high)


import types
from abc import abstractmethod
import numpy as np
from joblib import Parallel, delayed
import traceback
from collections.abc import Iterable
import warnings
from stopit import threading_timeoutable, TimeoutException
from tpot2.parent_selectors import survival_select_NSGA2
import time
import dask
import stopit
from dask.diagnostics import ProgressBar
from tqdm.dask import TqdmCallback
from dask.distributed import progress

import func_timeout

def process_scores(scores, n):
    '''
    Purpose: This function processes a list of scores to ensure that each score list has the same length, n. If a score list is shorter than n, the function fills the list with either "TIMEOUT" or "INVALID" values.

    Parameters:

        scores: A list of score lists. Each score list represents a set of scores for a particular player or team. The score lists may have different lengths.
        n: An integer representing the desired length for each score list.

    Returns:

        The scores list, after processing.
    
    '''
    for i in range(len(scores)):
        if len(scores[i]) < n:
            if "TIMEOUT" in scores[i]:
                scores[i] = ["TIMEOUT" for j in range(n)]
            else:
                scores[i] = ["INVALID" for j in range(n)]
    return scores


def objective_nan_wrapper(  individual, 
                            objective_function,
                            verbose=0,
                            timeout=None,
                            **objective_kwargs):
    with warnings.catch_warnings(record=True) as w:  #catches all warnings in w so it can be supressed by verbose                
        try:
            
            if timeout is None:
                value = objective_function(individual, **objective_kwargs)
            else:
                value = func_timeout.func_timeout(timeout, objective_function, args=[individual], kwargs=objective_kwargs)
            
            if not isinstance(value, Iterable):
                value = [value]               

            if len(w) and verbose>=4:
                
                warnings.warn(w[0].message)
            return value
        except func_timeout.exceptions.FunctionTimedOut:
            if verbose >= 4:
                print(f'WARNING AN INDIVIDUAL TIMED OUT')
            return ["TIMEOUT"]
        except Exception as e:
            if verbose == 4:
                print(f'WARNING THIS INDIVIDUAL CAUSED AND EXCEPTION \n {e}')
            if verbose >= 5:
                trace = traceback.format_exc()
                print(f'WARNING THIS INDIVIDUAL CAUSED AND EXCEPTION \n {e} \n {trace}')
            return ["INVALID"]
        

def eval_objective_list(ind, objective_list, verbose=0,**objective_kwargs):

    scores = np.concatenate([objective_nan_wrapper(ind, obj, verbose,**objective_kwargs) for obj in objective_list ])
    return scores

def parallel_eval_objective_list(individual_list,
                                objective_list,
                                n_jobs = 1,
                                verbose=0,
                                timeout=None,
                                n_expected_columns=None,
                                client=None,
                                **objective_kwargs):

    #offspring_scores = Parallel(n_jobs=n_jobs)(delayed(eval_objective_list)(ind,  objective_list, verbose, timeout=timeout)  for ind in individual_list )
    
    # delayed_values = [dask.delayed(eval_objective_list)(ind,  objective_list, verbose, timeout=timeout,**objective_kwargs)  for ind in individual_list]
    # with TqdmCallback(desc="Evaluating Individuals", disable=verbose<2, leave=False):
    #     offspring_scores = list(dask.compute( *delayed_values,
    #                             num_workers=n_jobs))
    # del delayed_values
    if client is None:
        client = dask.distributed.get_client()
    futures = [client.submit(eval_objective_list, ind,  objective_list, verbose, timeout=timeout,**objective_kwargs)  for ind in individual_list]
    
    if verbose >= 2:
        dask.distributed.progress(futures, notebook=False)
    
    dask.distributed.wait(futures)
    
    offspring_scores = []
    # todo optimize this
    for future in futures:
        if not future.exception():
            offspring_scores.append(future.result())
        else:
            offspring_scores.append(["INVALID"])

    if n_expected_columns is not None:
        offspring_scores = process_scores(offspring_scores, n_expected_columns)
    return offspring_scores


###################
# Parallel optimization
#############

@threading_timeoutable(np.nan) #TODO timeout behavior
def optimize_objective(ind, objective, steps=5, verbose=0):
    
    with warnings.catch_warnings(record=True) as w:  #catches all warnings in w so it can be supressed by verbose                
        try:
            value = ind.optimize(objective, steps=steps)
            if not isinstance(value, Iterable):
                value = [value]               

            if len(w) and verbose>=2:
                warnings.warn(w[0].message)
            return value
        except Exception as e:
            if verbose >= 2:
                print('WARNING THIS INDIVIDUAL CAUSED AND EXCEPTION')
                print(e)
                print()
            if verbose >= 3:
                print(traceback.format_exc())
                print()
            return [np.nan]



def parallel_optimize_objective(individual_list,
                                objective,
                                n_jobs = 1,
                                verbose=0,
                                steps=5,
                                timeout=None,
                                **objective_kwargs,  ):

    Parallel(n_jobs=n_jobs)(delayed(optimize_objective)(ind,  objective,  steps, verbose, timeout=timeout)  for ind in individual_list ) #TODO: parallelize









import tpot2
import numpy as np
from ..estimator import TPOTEstimator
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from tpot2.parent_selectors import survival_select_NSGA2, TournamentSelection_Dominated
#TODO These do not follow sklearn conventions of __init__

class TPOTRegressor(TPOTEstimator):
    def __init__(       self,
                        scorers=['neg_mean_squared_error'], 
                        scorers_weights=[1],
                        classification=False,
                        cv=5,
                        other_objective_functions=[tpot2.estimator_objective_functions.average_path_length_objective], #tpot2.estimator_objective_functions.number_of_nodes_objective],
                        other_objective_functions_weights=[-1],
                        objective_function_names=None,
                        bigger_is_better=True,
                        max_depth=np.inf,
                        max_size=np.inf, 
                        max_children=np.inf,
                        root_config_dict='Auto',
                        inner_config_dict=["selectors", "transformers"],
                        leaf_config_dict=None,                        
                        cross_val_predict_cv=0,
                        subsets=None,
                        memory=None,
                        preprocessing=False,  
                        validation_strategy="none",
                        validation_fraction=.2,
                        population_size=50,
                        initial_population_size=None,
                        population_scaling=.5, 
                        generations_until_end_population=1,  
                        generations=50,
                        early_stop=None,
                        scorers_early_stop_tol=0.001,
                        other_objectives_early_stop_tol=None,
                        max_time_seconds=float('inf'), 
                        max_eval_time_seconds=60*10, 
                        n_jobs=1,
                        memory_limit="4GB",
                        client=None,
                        survival_percentage=1,
                        crossover_probability=.2,
                        mutate_probability=.7,
                        mutate_then_crossover_probability=.05,
                        crossover_then_mutate_probability=.05,
                        survival_selector=survival_select_NSGA2,
                        parent_selector=TournamentSelection_Dominated,
                        budget_range=None,
                        budget_scaling=.5,
                        generations_until_end_budget=1,  
                        stepwise_steps=5,
                        threshold_evaluation_early_stop=None, 
                        threshold_evaluation_scaling=.5,
                        min_history_threshold=20,
                        selection_evaluation_early_stop=None, 
                        selection_evaluation_scaling=.5, 
                        n_initial_optimizations=0,
                        optimization_cv=3,
                        max_optimize_time_seconds=60*20,
                        optimization_steps=10,
                        warm_start=False,
                        subset_column=None,
                        evolver="nsga2",
                        verbose=0,
                        periodic_checkpoint_folder=None, 
                        callback: tpot2.CallBackInterface=None,
                        processes = True,
        ):
        super(TPOTRegressor,self).__init__(
            scorers=scorers, 
            scorers_weights=scorers_weights,
            classification=classification,
            cv=cv,
            other_objective_functions=other_objective_functions, #tpot2.estimator_objective_functions.number_of_nodes_objective],
            other_objective_functions_weights=other_objective_functions_weights,
            objective_function_names=objective_function_names,
            bigger_is_better=bigger_is_better,
            max_depth=max_depth,
            max_size=max_size, 
            max_children=max_children,
            root_config_dict=root_config_dict,
            inner_config_dict=inner_config_dict,
            leaf_config_dict=leaf_config_dict,                        
            cross_val_predict_cv=cross_val_predict_cv,
            subsets=subsets,
            memory=memory,
            preprocessing=preprocessing,  
            validation_strategy=validation_strategy,
            validation_fraction=validation_fraction,
            population_size=population_size,
            initial_population_size=initial_population_size,
            population_scaling=population_scaling, 
            generations_until_end_population=generations_until_end_population,  
            generations=generations,
            early_stop=early_stop,
            scorers_early_stop_tol=scorers_early_stop_tol,
            other_objectives_early_stop_tol=other_objectives_early_stop_tol,
            max_time_seconds=max_time_seconds, 
            max_eval_time_seconds=max_eval_time_seconds, 
            n_jobs=n_jobs,
            memory_limit=memory_limit,
            client=client,
            survival_percentage=survival_percentage,
            crossover_probability=crossover_probability,
            mutate_probability=mutate_probability,
            mutate_then_crossover_probability=mutate_then_crossover_probability,
            crossover_then_mutate_probability=crossover_then_mutate_probability,
            survival_selector=survival_selector,
            parent_selector=parent_selector,
            budget_range=budget_range,
            budget_scaling=budget_scaling,
            generations_until_end_budget=generations_until_end_budget,  
            stepwise_steps=stepwise_steps,
            threshold_evaluation_early_stop=threshold_evaluation_early_stop, 
            threshold_evaluation_scaling=threshold_evaluation_scaling,
            min_history_threshold=min_history_threshold,
            selection_evaluation_early_stop=selection_evaluation_early_stop, 
            selection_evaluation_scaling=selection_evaluation_scaling, 
            n_initial_optimizations=n_initial_optimizations,
            optimization_cv=optimization_cv,
            max_optimize_time_seconds=max_optimize_time_seconds,
            optimization_steps=optimization_steps,
            warm_start=warm_start,
            subset_column=subset_column,
            evolver=evolver,
            verbose=verbose,
            periodic_checkpoint_folder=periodic_checkpoint_folder, 
            callback=callback,
            processes=processes,
)


class TPOTClassifier(TPOTEstimator):
    def __init__(       self,
                        scorers=['roc_auc_ovr'], 
                        scorers_weights=[1],
                        classification=True,
                        cv=5,
                        other_objective_functions=[tpot2.estimator_objective_functions.average_path_length_objective], #tpot2.estimator_objective_functions.number_of_nodes_objective],
                        other_objective_functions_weights=[-1],
                        objective_function_names=None,
                        bigger_is_better=True,
                        max_depth=np.inf,
                        max_size=np.inf, 
                        max_children=np.inf,
                        root_config_dict='Auto',
                        inner_config_dict=["selectors", "transformers"],
                        leaf_config_dict=None,                        
                        cross_val_predict_cv=0,
                        subsets=None,
                        memory=None,
                        preprocessing=False,  
                        validation_strategy="none",
                        validation_fraction=.2,
                        population_size=50,
                        initial_population_size=None,
                        population_scaling=.5, 
                        generations_until_end_population=1,  
                        generations=50,
                        early_stop=None,
                        scorers_early_stop_tol=0.001,
                        other_objectives_early_stop_tol=None,
                        max_time_seconds=float('inf'), 
                        max_eval_time_seconds=60*10, 
                        n_jobs=1,
                        memory_limit="4GB",
                        client=None,
                        survival_percentage=1,
                        crossover_probability=.2,
                        mutate_probability=.7,
                        mutate_then_crossover_probability=.05,
                        crossover_then_mutate_probability=.05,
                        survival_selector=survival_select_NSGA2,
                        parent_selector=TournamentSelection_Dominated,
                        budget_range=None,
                        budget_scaling=.5,
                        generations_until_end_budget=1,  
                        stepwise_steps=5,
                        threshold_evaluation_early_stop=None, 
                        threshold_evaluation_scaling=.5,
                        min_history_threshold=20,
                        selection_evaluation_early_stop=None, 
                        selection_evaluation_scaling=.5, 
                        n_initial_optimizations=0,
                        optimization_cv=3,
                        max_optimize_time_seconds=60*20,
                        optimization_steps=10,
                        warm_start=False,
                        subset_column=None,
                        evolver="nsga2",
                        verbose=0,
                        periodic_checkpoint_folder=None, 
                        callback: tpot2.CallBackInterface=None,
                        processes = True,
        ):
        super(TPOTClassifier,self).__init__(
                        scorers=scorers, 
            scorers_weights=scorers_weights,
            classification=classification,
            cv=cv,
            other_objective_functions=other_objective_functions, #tpot2.estimator_objective_functions.number_of_nodes_objective],
            other_objective_functions_weights=other_objective_functions_weights,
            objective_function_names=objective_function_names,
            bigger_is_better=bigger_is_better,
            max_depth=max_depth,
            max_size=max_size, 
            max_children=max_children,
            root_config_dict=root_config_dict,
            inner_config_dict=inner_config_dict,
            leaf_config_dict=leaf_config_dict,                        
            cross_val_predict_cv=cross_val_predict_cv,
            subsets=subsets,
            memory=memory,
            preprocessing=preprocessing,  
            validation_strategy=validation_strategy,
            validation_fraction=validation_fraction,
            population_size=population_size,
            initial_population_size=initial_population_size,
            population_scaling=population_scaling, 
            generations_until_end_population=generations_until_end_population,  
            generations=generations,
            early_stop=early_stop,
            scorers_early_stop_tol=scorers_early_stop_tol,
            other_objectives_early_stop_tol=other_objectives_early_stop_tol,
            max_time_seconds=max_time_seconds, 
            max_eval_time_seconds=max_eval_time_seconds, 
            n_jobs=n_jobs,
            memory_limit=memory_limit,
            client=client,
            survival_percentage=survival_percentage,
            crossover_probability=crossover_probability,
            mutate_probability=mutate_probability,
            mutate_then_crossover_probability=mutate_then_crossover_probability,
            crossover_then_mutate_probability=crossover_then_mutate_probability,
            survival_selector=survival_selector,
            parent_selector=parent_selector,
            budget_range=budget_range,
            budget_scaling=budget_scaling,
            generations_until_end_budget=generations_until_end_budget,  
            stepwise_steps=stepwise_steps,
            threshold_evaluation_early_stop=threshold_evaluation_early_stop, 
            threshold_evaluation_scaling=threshold_evaluation_scaling,
            min_history_threshold=min_history_threshold,
            selection_evaluation_early_stop=selection_evaluation_early_stop, 
            selection_evaluation_scaling=selection_evaluation_scaling, 
            n_initial_optimizations=n_initial_optimizations,
            optimization_cv=optimization_cv,
            max_optimize_time_seconds=max_optimize_time_seconds,
            optimization_steps=optimization_steps,
            warm_start=warm_start,
            subset_column=subset_column,
            evolver=evolver,
            verbose=verbose,
            periodic_checkpoint_folder=periodic_checkpoint_folder, 
            callback=callback,
            processes = processes,
        )


    def predict(self, X, **predict_params):
        check_is_fitted(self)
        #X=check_array(X)
        return self.fitted_pipeline_.predict(X,**predict_params)


from sklearn.base import BaseEstimator
from sklearn.utils.metaestimators import available_if
import numpy as np
import typing
import sklearn.metrics
import tpot2.estimator_objective_functions
from functools import partial
import tpot2.config
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from tpot2.parent_selectors import survival_select_NSGA2, TournamentSelection_Dominated
from sklearn.preprocessing import LabelEncoder 
from sklearn.utils.multiclass import unique_labels 
import pandas as pd
from sklearn.model_selection import train_test_split
import tpot2
import distributed
from dask.distributed import Client
from dask.distributed import LocalCluster
import math

EVOLVERS = {"nsga2":tpot2.BaseEvolver}



#TODO inherit from _BaseComposition?
class TPOTEstimator(BaseEstimator):
    def __init__(self,  scorers, 
                        scorers_weights,
                        classification,
                        cv = 5,
                        other_objective_functions=[tpot2.estimator_objective_functions.average_path_length_objective], #tpot2.estimator_objective_functions.number_of_nodes_objective],
                        other_objective_functions_weights = [-1],
                        objective_function_names = None,
                        bigger_is_better = True,
                        max_depth = np.inf,
                        max_size = np.inf, 
                        max_children = np.inf,
                        root_config_dict= 'Auto',
                        inner_config_dict=["selectors", "transformers"],
                        leaf_config_dict= None,                        
                        cross_val_predict_cv = 0,
                        subsets = None,
                        memory = None,
                        preprocessing = False,  
                        validation_strategy = "none",
                        validation_fraction = .2,
                        population_size = 50,
                        initial_population_size = None,
                        population_scaling = .5, 
                        generations_until_end_population = 1,  
                        generations = 50,
                        early_stop = None,
                        scorers_early_stop_tol = 0.001,
                        other_objectives_early_stop_tol =None,
                        max_time_seconds=float('inf'), 
                        max_eval_time_seconds=60*10, 
                        n_jobs=1,
                        memory_limit = "4GB",
                        client = None,
                        survival_percentage = 1,
                        crossover_probability=.2,
                        mutate_probability=.7,
                        mutate_then_crossover_probability=.05,
                        crossover_then_mutate_probability=.05,
                        survival_selector = survival_select_NSGA2,
                        parent_selector = TournamentSelection_Dominated,
                        budget_range = None,
                        budget_scaling = .5,
                        generations_until_end_budget = 1,  
                        stepwise_steps = 5,
                        threshold_evaluation_early_stop = None, 
                        threshold_evaluation_scaling = .5,
                        min_history_threshold = 20,
                        selection_evaluation_early_stop = None, 
                        selection_evaluation_scaling = .5, 
                        n_initial_optimizations = 0,
                        optimization_cv = 3,
                        max_optimize_time_seconds=60*20,
                        optimization_steps = 10,
                        warm_start = False,
                        subset_column = None,
                        evolver = "nsga2",
                        verbose = 0,
                        periodic_checkpoint_folder = None, 
                        callback: tpot2.CallBackInterface = None,
                        processes = True,
                        ):
                        
        '''
        An sklearn baseestimator that uses genetic programming to optimize a pipeline.
        
        Parameters
        ----------
        
        scorers : (list, scorer)
            A scorer or list of scorers to be used in the cross-validation process. 
            see https://scikit-learn.org/stable/modules/model_evaluation.html
        
        scorers_weights : list
            A list of weights to be applied to the scorers during the optimization process.
        
        classification : bool
            If True, the problem is treated as a classification problem. If False, the problem is treated as a regression problem.
            Used to determine the CV strategy.
        
        cv : int, cross-validator
            - (int): Number of folds to use in the cross-validation process. By uses the sklearn.model_selection.KFold cross-validator for regression and StratifiedKFold for classification. In both cases, shuffled is set to True.
            - (sklearn.model_selection.BaseCrossValidator): A cross-validator to use in the cross-validation process.
                - max_depth (int): The maximum depth from any node to the root of the pipelines to be generated.
        
        other_objective_functions : list, default=[tpot2.estimator_objective_functions.average_path_length_objective]
            A list of other objective functions to apply to the pipeline.
        
        other_objective_functions_weights : list, default=[-1]
            A list of weights to be applied to the other objective functions.
        
        objective_function_names : list, default=None
            A list of names to be applied to the objective functions. If None, will use the names of the objective functions.
        
        bigger_is_better : bool, default=True
            If True, the objective function is maximized. If False, the objective function is minimized. Use negative weights to reverse the direction.

        max_depth : int, default=np.inf
            The maximum depth from any node to the root of the pipelines to be generated.
        
        max_size : int, default=np.inf
            The maximum number of nodes of the pipelines to be generated.
        
        max_children : ind, default=np.inf 
            The maximum number of children nodes in the pipelines can have. If set to 1, the pipelines will be linear.
        
        root_config_dict : dict, default='auto'
            The configuration dictionary to use for the root node of the model.
            If 'auto', will use "classifiers" if classification=True, else "regressors".
            - 'selectors' : A selection of sklearn Selector methods.
            - 'classifiers' : A selection of sklearn Classifier methods.
            - 'regressors' : A selection of sklearn Regressor methods.
            - 'transformers' : A selection of sklearn Transformer methods.
            - 'arithmetic_transformer' : A selection of sklearn Arithmetic Transformer methods that replicate symbolic classification/regression operators.
            - 'passthrough' : A node that just passes though the input. Useful for passing through raw inputs into inner nodes.
            - 'feature_set_selector' : A selector that pulls out specific subsets of columns from the data. Only well defined as a leaf.
                                        Subsets are set with the subsets parameter.
            - 'skrebate' : Includes ReliefF, SURF, SURFstar, MultiSURF.
            - 'MDR' : Includes MDR.
            - 'ContinuousMDR' : Includes ContinuousMDR.
            - 'genetic encoders' : Includes Genetic Encoder methods as used in AutoQTL.
            - 'FeatureEncodingFrequencySelector': Includes FeatureEncodingFrequencySelector method as used in AutoQTL.
            - list : a list of strings out of the above options to include the corresponding methods in the configuration dictionary.
        
        inner_config_dict : dict, default=["selectors", "transformers"]
            The configuration dictionary to use for the inner nodes of the model generation.
            Default ["selectors", "transformers"]
            - 'selectors' : A selection of sklearn Selector methods.
            - 'classifiers' : A selection of sklearn Classifier methods.
            - 'regressors' : A selection of sklearn Regressor methods.
            - 'transformers' : A selection of sklearn Transformer methods.
            - 'arithmetic_transformer' : A selection of sklearn Arithmetic Transformer methods that replicate symbolic classification/regression operators.
            - 'passthrough' : A node that just passes though the input. Useful for passing through raw inputs into inner nodes.
            - 'feature_set_selector' : A selector that pulls out specific subsets of columns from the data. Only well defined as a leaf.
                                        Subsets are set with the subsets parameter.
            - 'skrebate' : Includes ReliefF, SURF, SURFstar, MultiSURF.
            - 'MDR' : Includes MDR.
            - 'ContinuousMDR' : Includes ContinuousMDR.
            - 'genetic encoders' : Includes Genetic Encoder methods as used in AutoQTL.
            - 'FeatureEncodingFrequencySelector': Includes FeatureEncodingFrequencySelector method as used in AutoQTL.
            - list : a list of strings out of the above options to include the corresponding methods in the configuration dictionary.
            - None : If None and max_depth>1, the root_config_dict will be used for the inner nodes as well.
        
        leaf_config_dict : dict, default=None 
            The configuration dictionary to use for the leaf node of the model. If set, leaf nodes must be from this dictionary.
            Otherwise leaf nodes will be generated from the root_config_dict. 
            Default None
            - 'selectors' : A selection of sklearn Selector methods.
            - 'classifiers' : A selection of sklearn Classifier methods.
            - 'regressors' : A selection of sklearn Regressor methods.
            - 'transformers' : A selection of sklearn Transformer methods.
            - 'arithmetic_transformer' : A selection of sklearn Arithmetic Transformer methods that replicate symbolic classification/regression operators.
            - 'passthrough' : A node that just passes though the input. Useful for passing through raw inputs into inner nodes.
            - 'feature_set_selector' : A selector that pulls out specific subsets of columns from the data. Only well defined as a leaf.
                                        Subsets are set with the subsets parameter.
            - 'skrebate' : Includes ReliefF, SURF, SURFstar, MultiSURF.
            - 'MDR' : Includes MDR.
            - 'ContinuousMDR' : Includes ContinuousMDR.
            - 'genetic encoders' : Includes Genetic Encoder methods as used in AutoQTL.
            - 'FeatureEncodingFrequencySelector': Includes FeatureEncodingFrequencySelector method as used in AutoQTL.
            - list : a list of strings out of the above options to include the corresponding methods in the configuration dictionary.
            - None : If None, a leaf will not be required (i.e. the pipeline can be a single root node). Leaf nodes will be generated from the inner_config_dict.
        
        cross_val_predict_cv : int, default=0
            Number of folds to use for the cross_val_predict function for inner classifiers and regressors. Estimators will still be fit on the full dataset, but the following node will get the outputs from cross_val_predict.
            
            - 0-1 : When set to 0 or 1, the cross_val_predict function will not be used. The next layer will get the outputs from fitting and transforming the full dataset.
            - >=2 : When fitting pipelines with inner classifiers or regressors, they will still be fit on the full dataset. 
                    However, the output to the next node will come from cross_val_predict with the specified number of folds.
         
        subsets : str or list, default=None
            Sets the subsets that the FeatureSetSeletor will select from if set as an option in one of the configuration dictionaries.
            - str : If a string, it is assumed to be a path to a csv file with the subsets. 
                The first column is assumed to be the name of the subset and the remaining columns are the features in the subset.
            - list or np.ndarray : If a list or np.ndarray, it is assumed to be a list of subsets.
            - None : If None, each column will be treated as a subset. One column will be selected per subset.
            If subsets is None, each column will be treated as a subset. One column will be selected per subset.


        memory: Memory object or string, default=None
            If supplied, pipeline will cache each transformer after calling fit. This feature
            is used to avoid computing the fit transformers within a pipeline if the parameters
            and input data are identical with another fitted pipeline during optimization process.
            - String 'auto':
                TPOT uses memory caching with a temporary directory and cleans it up upon shutdown.
            - String path of a caching directory
                TPOT uses memory caching with the provided directory and TPOT does NOT clean
                the caching directory up upon shutdown. If the directory does not exist, TPOT will
                create it.
            - Memory object:
                TPOT uses the instance of joblib.Memory for memory caching,
                and TPOT does NOT clean the caching directory up upon shutdown.
            - None:
                TPOT does not use memory caching.

        preprocessing : bool or BaseEstimator/Pipeline, 
            EXPERIMENTAL
            A pipeline that will be used to preprocess the data before CV.
            - bool : If True, will use a default preprocessing pipeline.
            - Pipeline : If an instance of a pipeline is given, will use that pipeline as the preprocessing pipeline.
              
        validation_strategy : str, default='none'
            EXPERIMENTAL The validation strategy to use for selecting the final pipeline from the population. TPOT2 may overfit the cross validation score. A second validation set can be used to select the final pipeline.
            - 'auto' : Automatically determine the validation strategy based on the dataset shape.
            - 'reshuffled' : Use the same data for cross validation and final validation, but with different splits for the folds. This is the default for small datasets. 
            - 'split' : Use a separate validation set for final validation. Data will be split according to validation_fraction. This is the default for medium datasets. 
            - 'none' : Do not use a separate validation set for final validation. Select based on the original cross-validation score. This is the default for large datasets.

        validation_fraction : float, default=0.2
          EXPERIMENTAL The fraction of the dataset to use for the validation set when validation_strategy is 'split'. Must be between 0 and 1.
        
        population_size : int, default=50
            Size of the population
        
        initial_population_size : int, default=None
            Size of the initial population. If None, population_size will be used.
        
        population_scaling : int, default=0.5
            Scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
        
        generations_until_end_population : int, default=1  
            Number of generations until the population size reaches population_size            
        
        generations : int, default=50
            Number of generations to run
        
        early_stop : int, default=None
            Number of generations without improvement before early stopping. All objectives must have converged within the tolerance for this to be triggered.
        
        scorers_early_stop_tol : 
            -list of floats
                list of tolerances for each scorer. If the difference between the best score and the current score is less than the tolerance, the individual is considered to have converged
                If an index of the list is None, that item will not be used for early stopping
            -int 
                If an int is given, it will be used as the tolerance for all objectives
        
        other_objectives_early_stop_tol : 
            -list of floats
                list of tolerances for each of the other objective function. If the difference between the best score and the current score is less than the tolerance, the individual is considered to have converged
                If an index of the list is None, that item will not be used for early stopping
            -int 
                If an int is given, it will be used as the tolerance for all objectives
    
        max_time_seconds : float, default=float("inf")
            Maximum time to run the optimization. If none or inf, will run until the end of the generations.
        
        max_eval_time_seconds : float, default=60*5
            Maximum time to evaluate a single individual. If none or inf, there will be no time limit per evaluation.
        
        n_jobs : int, default=1
            Number of processes to run in parallel.
        
        memory_limit : str, default="4GB"
            Memory limit for each job. See Dask [LocalCluster documentation](https://distributed.dask.org/en/stable/api.html#distributed.Client) for more information.
        
        client : dask.distributed.Client, default=None
            A dask client to use for parallelization. If not None, this will override the n_jobs and memory_limit parameters. If None, will create a new client with num_workers=n_jobs and memory_limit=memory_limit. 
        
        survival_percentage : float, default=1
            Percentage of the population size to utilize for mutation and crossover at the beginning of the generation. The rest are discarded. Individuals are selected with the selector passed into survival_selector. The value of this parameter must be between 0 and 1, inclusive. 
            For example, if the population size is 100 and the survival percentage is .5, 50 individuals will be selected with NSGA2 from the existing population. These will be used for mutation and crossover to generate the next 100 individuals for the next generation. The remainder are discarded from the live population. In the next generation, there will now be the 50 parents + the 100 individuals for a total of 150. Surivival percentage is based of the population size parameter and not the existing population size. Therefore, in the next generation we will still select 50 individuals from the currently existing 150.
        
        crossover_probability : float, default=.2
            Probability of generating a new individual by crossover between two individuals.
        
        mutate_probability : float, default=.7
            Probability of generating a new individual by crossover between one individuals.
        
        mutate_then_crossover_probability : float, default=.05
            Probability of generating a new individual by mutating two individuals followed by crossover.
        
        crossover_then_mutate_probability : float, default=.05
            Probability of generating a new individual by crossover between two individuals followed by a mutation of the resulting individual.
        
        n_parents : int, default=2
            Number of parents to use for crossover. Must be greater than 1.
        
        survival_selector : function, default=survival_select_NSGA2
            Function to use to select individuals for survival. Must take a matrix of scores and return selected indexes.
            Used to selected population_size * survival_percentage individuals at the start of each generation to use for mutation and crossover.
        
        parent_selector : function, default=parent_select_NSGA2
            Function to use to select pairs parents for crossover and individuals for mutation. Must take a matrix of scores and return selected indexes.
        
        budget_range : list [start, end], default=None
            A starting and ending budget to use for the budget scaling.
        
        budget_scaling float : [0,1], default=0.5
            A scaling factor to use when determining how fast we move the budget from the start to end budget.
        
        generations_until_end_budget : int, default=1
            The number of generations to run before reaching the max budget.
        
        stepwise_steps : int, default=1
            The number of staircase steps to take when scaling the budget and population size.
        
        threshold_evaluation_early_stop : list [start, end], default=None
            starting and ending percentile to use as a threshold for the evaluation early stopping.
            Values between 0 and 100.
        
        threshold_evaluation_scaling : float [0,inf), default=0.5
            A scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
            Must be greater than zero. Higher numbers will move the threshold to the end faster.
        
        min_history_threshold : int, default=0
            The minimum number of previous scores needed before using threshold early stopping.
        
        selection_evaluation_early_stop : list, default=None
            A lower and upper percent of the population size to select each round of CV.
            Values between 0 and 1.
        
        selection_evaluation_scaling : float, default=0.5 
            A scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
            Must be greater than zero. Higher numbers will move the threshold to the end faster.
        
        n_initial_optimizations : int, default=0
            Number of individuals to optimize before starting the evolution.
        
        optimization_cv : int 
           Number of folds to use for the optuna optimization's internal cross-validation.
        
        max_optimize_time_seconds : float, default=60*5
            Maximum time to run an optimization
        
        optimization_steps : int, default=10
            Number of steps per optimization
          
        warm_start : bool, default=False
            If True, will use the continue the evolutionary algorithm from the last generation of the previous run.
         
        subset_column : str or int, default=None
            EXPERIMENTAL The column to use for the subset selection. Must also pass in unique_subset_values to GraphIndividual to function.
         
        evolver : tpot2.evolutionary_algorithms.eaNSGA2.eaNSGA2_Evolver), default=eaNSGA2_Evolver
            The evolver to use for the optimization process. See tpot2.evolutionary_algorithms
            - type : an type or subclass of a BaseEvolver
            - "nsga2" : tpot2.evolutionary_algorithms.eaNSGA2.eaNSGA2_Evolver
        
        verbose : int, default=1 
            How much information to print during the optimization process. Higher values include the information from lower values.
            0. nothing
            1. progress bar
            2. evaluations progress bar
            3. best individual
            4. warnings
            >=5. full warnings trace
        
        periodic_checkpoint_folder : str, default=None
            Folder to save the population to periodically. If None, no periodic saving will be done.
            If provided, training will resume from this checkpoint.
        
        callback : tpot2.CallBackInterface, default=None
            Callback object. Not implemented

        processes : bool, default=True
            If True, will use multiprocessing to parallelize the optimization process. If False, will use threading.
            True seems to perform better. However, False is required for interactive debugging.
            
        '''

        # sklearn BaseEstimator must have a corresponding attribute for each parameter.
        # These should not be modified once set.
        self.scorers = scorers
        self.scorers_weights = scorers_weights
        self.classification = classification
        self.population_size = population_size
        self.generations = generations
        self.initial_population_size = initial_population_size
        self.population_scaling = population_scaling
        self.generations_until_end_population = generations_until_end_population
        self.callback = callback
        self.n_jobs= n_jobs
        self.cv = cv
        self.verbose = verbose
        self.other_objective_functions = other_objective_functions
        self.other_objective_functions_weights = other_objective_functions_weights
        self.bigger_is_better = bigger_is_better
        self.evolver = evolver

        self.max_depth = max_depth
        self.max_size = max_size
        self.max_children = max_children
        self.root_config_dict= root_config_dict
        self.inner_config_dict= inner_config_dict
        self.leaf_config_dict= leaf_config_dict
        self.subsets = subsets
        self.max_time_seconds = max_time_seconds 
        self.max_eval_time_seconds = max_eval_time_seconds
        self.memory_limit = memory_limit
        self.n_initial_optimizations  = n_initial_optimizations  
        self.optimization_cv  = optimization_cv
        self.max_optimize_time_seconds = max_optimize_time_seconds 
        self.optimization_steps = optimization_steps 
        self.periodic_checkpoint_folder = periodic_checkpoint_folder
        self.threshold_evaluation_early_stop =threshold_evaluation_early_stop
        self.threshold_evaluation_scaling =  threshold_evaluation_scaling
        self.min_history_threshold = min_history_threshold
        self.selection_evaluation_early_stop = selection_evaluation_early_stop
        self.selection_evaluation_scaling =  selection_evaluation_scaling
        self.scorers_early_stop_tol = scorers_early_stop_tol
        self.other_objectives_early_stop_tol = other_objectives_early_stop_tol
        self.early_stop = early_stop
        self.warm_start = warm_start
        self.memory = memory
        self.cross_val_predict_cv = cross_val_predict_cv
        self.budget_range = budget_range
        self.budget_scaling = budget_scaling
        self.generations_until_end_budget = generations_until_end_budget
        self.preprocessing = preprocessing
        self.validation_strategy = validation_strategy
        self.validation_fraction = validation_fraction
        self.subset_column = subset_column
        self.stepwise_steps = stepwise_steps
        self.client = client

        self.survival_selector=survival_selector
        self.parent_selector=parent_selector
        self.survival_percentage = survival_percentage
        self.crossover_probability = crossover_probability
        self.mutate_probability = mutate_probability
        self.mutate_then_crossover_probability= mutate_then_crossover_probability
        self.crossover_then_mutate_probability= crossover_then_mutate_probability

        self.objective_function_names = objective_function_names

        self.processes = processes

        #Initialize other used params


        if self.initial_population_size is None:
            self._initial_population_size = self.population_size
        else:
            self._initial_population_size = self.initial_population_size

        if isinstance(self.scorers, str):
            self._scorers = [self.scorers]

        elif callable(self.scorers):
            self._scorers = [self.scorers]
        else:
            self._scorers = self.scorers
        
        self._scorers = [sklearn.metrics.get_scorer(scoring) for scoring in self._scorers]
        self._scorers_early_stop_tol = self.scorers_early_stop_tol
        
        if isinstance(self.evolver, str):
            self._evolver = EVOLVERS[self.evolver]
        else:
            self._evolver = self.evolver
        
       

        self.objective_function_weights = [*scorers_weights, *other_objective_functions_weights]
        

        if self.objective_function_names is None:
            obj_names = [f.__name__ for f in other_objective_functions]
        else:
            obj_names = self.objective_function_names
        self.objective_names = [f._score_func.__name__ if hasattr(f,"_score_func") else f.__name__ for f in self._scorers] + obj_names
        
        
        if not isinstance(self.other_objectives_early_stop_tol, list):
            self._other_objectives_early_stop_tol = [self.other_objectives_early_stop_tol for _ in range(len(self.other_objective_functions))]
        else:
            self._other_objectives_early_stop_tol = self.other_objectives_early_stop_tol

        if not isinstance(self._scorers_early_stop_tol, list):
            self._scorers_early_stop_tol = [self._scorers_early_stop_tol for _ in range(len(self._scorers))]
        else:
            self._scorers_early_stop_tol = self._scorers_early_stop_tol

        self.early_stop_tol = [*self._scorers_early_stop_tol, *self._other_objectives_early_stop_tol]
        
        self._evolver_instance = None
        self.evaluated_individuals = None


    def fit(self, X, y):
        if self.client is not None: #If user passed in a client manually
           _client = self.client
        else:

            if self.verbose >= 4:
                silence_logs = 30
            elif self.verbose >=5:
                silence_logs = 40
            else:
                silence_logs = 50
            cluster = LocalCluster(n_workers=self.n_jobs, #if no client is passed in and no global client exists, create our own
                    threads_per_worker=1,
                    processes=self.processes,
                    silence_logs=silence_logs,
                    memory_limit=self.memory_limit)
            _client = Client(cluster)


        self.evaluated_individuals = None
        #determine validation strategy
        if self.validation_strategy == 'auto':
            nrows = X.shape[0]
            ncols = X.shape[1]

            if nrows/ncols < 20:
                validation_strategy = 'reshuffled'
            elif nrows/ncols < 100:
                validation_strategy = 'split'
            else:
                validation_strategy = 'none'
        else:
            validation_strategy = self.validation_strategy

        if validation_strategy == 'split':
            if self.classification:
                X, X_val, y, y_val = train_test_split(X, y, test_size=self.validation_fraction, stratify=y, random_state=42)
            else:
                X, X_val, y, y_val = train_test_split(X, y, test_size=self.validation_fraction, random_state=42)


        X_original = X
        if self.preprocessing:
            X = pd.DataFrame(X)
            self._preprocessing_pipeline = sklearn.pipeline.make_pipeline(tpot2.CatImpute(), tpot2.NumericImpute(), tpot2.CatOneHotEncoder(),sklearn.preprocessing.StandardScaler())
            X = self._preprocessing_pipeline.fit_transform(X)
        else:
            self._preprocessing_pipeline = None

        #_, y = sklearn.utils.check_X_y(X, y, y_numeric=True)

        #Set up the configuation dictionaries and the search spaces

        if isinstance(self.cv, int) or isinstance(self.cv, float):
            n_folds = self.cv
        else:
            n_folds = self.cv.n_splits

        n_samples= int(math.floor(X.shape[0]/n_folds))
        n_features=X.shape[1]

        if isinstance(X, pd.DataFrame):
            self.feature_names = X.columns
        else:
            self.feature_names = None

        if self.root_config_dict == 'Auto':
            if self.classification:
                root_config_dict = get_configuration_dictionary("classifiers", n_samples, n_features, self.classification, subsets=self.subsets, feature_names=self.feature_names)
            else:
                root_config_dict = get_configuration_dictionary("regressors", n_samples, n_features, self.classification,subsets=self.subsets, feature_names=self.feature_names)
        else:
            root_config_dict = get_configuration_dictionary(self.root_config_dict, n_samples, n_features, self.classification, subsets=self.subsets,feature_names=self.feature_names)

        inner_config_dict = get_configuration_dictionary(self.inner_config_dict, n_samples, n_features, self.classification,subsets=self.subsets, feature_names=self.feature_names)
        leaf_config_dict = get_configuration_dictionary(self.leaf_config_dict, n_samples, n_features, self.classification, subsets=self.subsets, feature_names=self.feature_names)

        if self.n_initial_optimizations > 0:
            #tmp = partial(tpot2.estimator_objective_functions.cross_val_score_objective,scorers= self._scorers, cv=self.optimization_cv, memory=self.memory, cross_val_predict_cv=self.cross_val_predict_cv, subset_column=self.subset_column )
            # optuna_objective = lambda ind,  X=X, y=y , scorers= self._scorers, cv=self.optimization_cv, memory=self.memory, cross_val_predict_cv=self.cross_val_predict_cv, subset_column=self.subset_column: tpot2.estimator_objective_functions.cross_val_score_objective(
            #     ind, 
            #     X=X, y=y, scorers= scorers, cv=cv, memory=memory, cross_val_predict_cv=cross_val_predict_cv, subset_column=subset_column )
            
            optuna_objective = partial(tpot2.estimator_objective_functions.cross_val_score_objective, X=X, y=y , scorers= self._scorers, cv=self.optimization_cv, memory=self.memory, cross_val_predict_cv=self.cross_val_predict_cv, subset_column=self.subset_column )
        else:
            optuna_objective = None


        #check if self.cv is a number
        if isinstance(self.cv, int) or isinstance(self.cv, float):
            if self.classification:
                self.cv_gen = sklearn.model_selection.StratifiedKFold(n_splits=self.cv, shuffle=True, random_state=42)
            else:
                self.cv_gen = sklearn.model_selection.KFold(n_splits=self.cv, shuffle=True, random_state=42)

        else:
            self.cv_gen = sklearn.model_selection.check_cv(self.cv, y, classifier=self.classification)


        self.individual_generator_instance = tpot2.estimator_graph_individual_generator(   
                                                            inner_config_dict=inner_config_dict,
                                                            root_config_dict=root_config_dict,
                                                            leaf_config_dict=leaf_config_dict,
                                                            max_depth = self.max_depth,
                                                            max_size = self.max_size,
                                                            max_children = self.max_children,
                                                                )

        if self.threshold_evaluation_early_stop is not None or self.selection_evaluation_early_stop is not None:
            evaluation_early_stop_steps = self.cv
        else:
            evaluation_early_stop_steps = None


        X_future = _client.scatter(X)
        y_future = _client.scatter(y)

        #.export_pipeline(memory=self.memory, cross_val_predict_cv=self.cross_val_predict_cv, subset_column=self.subset_column),
        #tmp = partial(objective_function_generator, scorers= self._scorers, cv=self.cv_gen, other_objective_functions=self.other_objective_functions )
        self.final_object_function_list =[ lambda pipeline_individual, X, y,is_classification=self.classification,
                scorers= self._scorers, cv=self.cv_gen, other_objective_functions=self.other_objective_functions,
                 memory=self.memory, cross_val_predict_cv=self.cross_val_predict_cv, subset_column=self.subset_column, **kwargs: objective_function_generator(
                                pipeline_individual,
                                #ind,
                                X, y, 
                                is_classification=is_classification,
                                scorers= scorers, cv=cv, other_objective_functions=other_objective_functions,
                                memory=memory, cross_val_predict_cv=cross_val_predict_cv, subset_column=subset_column,
                                **kwargs,
                                )]


        #If warm start and we have an evolver instance, use the existing one
        if not(self.warm_start and self._evolver_instance is not None):
            self._evolver_instance = self._evolver(   individual_generator=self.individual_generator_instance, 
                                            objective_functions=self.final_object_function_list,
                                            objective_function_weights = self.objective_function_weights,
                                            objective_names=self.objective_names,
                                            bigger_is_better = self.bigger_is_better,
                                            population_size= self.population_size,
                                            generations=self.generations,
                                            initial_population_size = self._initial_population_size,
                                            n_jobs=self.n_jobs,
                                            verbose = self.verbose,
                                            max_time_seconds =      self.max_time_seconds ,
                                            max_eval_time_seconds = self.max_eval_time_seconds,
                                            optimization_objective=optuna_objective,
                                            periodic_checkpoint_folder = self.periodic_checkpoint_folder,
                                            threshold_evaluation_early_stop = self.threshold_evaluation_early_stop,
                                            threshold_evaluation_scaling =  self.threshold_evaluation_scaling,
                                            min_history_threshold = self.min_history_threshold,

                                            selection_evaluation_early_stop = self.selection_evaluation_early_stop,
                                            selection_evaluation_scaling =  self.selection_evaluation_scaling,
                                            evaluation_early_stop_steps = evaluation_early_stop_steps,

                                            early_stop_tol = self.early_stop_tol,
                                            early_stop= self.early_stop,
                                            
                                            budget_range = self.budget_range,
                                            budget_scaling = self.budget_scaling,
                                            generations_until_end_budget = self.generations_until_end_budget,

                                            population_scaling = self.population_scaling,
                                            generations_until_end_population = self.generations_until_end_population,
                                            stepwise_steps = self.stepwise_steps,
                                            client = _client,
                                            objective_kwargs = {"X": X_future, "y": y_future},
                                            survival_selector=self.survival_selector,
                                            parent_selector=self.parent_selector,
                                            survival_percentage = self.survival_percentage,
                                            crossover_probability = self.crossover_probability,
                                            mutate_probability = self.mutate_probability,
                                            mutate_then_crossover_probability= self.mutate_then_crossover_probability,
                                            crossover_then_mutate_probability= self.crossover_then_mutate_probability,
                                            
                                            )

        
        self._evolver_instance.optimize()
        #self._evolver_instance.population.update_pareto_fronts(self.objective_names, self.objective_function_weights)
        #self.make_evaluated_individuals()


        self.evaluated_individuals = self._evolver_instance.population.evaluated_individuals.copy()
        self.evaluated_individuals = get_pareto_front(self.evaluated_individuals, self.objective_names, self.objective_function_weights)

        if validation_strategy == 'reshuffled':
            best_pareto_front_idx = list(self.pareto_front.index)
            best_pareto_front = self.pareto_front.loc[best_pareto_front_idx]['Instance']
            
            #reshuffle rows
            X, y = sklearn.utils.shuffle(X, y, random_state=1)
            X_future = _client.scatter(X)
            y_future = _client.scatter(y)

            val_objective_function_list = [lambda ind, X, y, is_classification=self.classification,scorers= self._scorers, cv=self.cv_gen, other_objective_functions=self.other_objective_functions, **kwargs: objective_function_generator(
                                                                                                ind,
                                                                                                X,y, 
                                                                                                is_classification=is_classification,
                                                                                                scorers= scorers, cv=cv, other_objective_functions=other_objective_functions,
                                                                                                **kwargs,
                                                                                                )]
            
            val_scores = tpot2.objectives.parallel_eval_objective_list(
                best_pareto_front,
                val_objective_function_list, n_jobs=self.n_jobs, verbose=self.verbose, timeout=self.max_eval_time_seconds,n_expected_columns=len(self.objective_names), client=_client, X= X_future, y= y_future)

            val_objective_names = ['validation_'+name for name in self.objective_names]
            self.objective_names_for_selection = val_objective_names
            self.evaluated_individuals.loc[best_pareto_front_idx,val_objective_names] = val_scores

        elif validation_strategy == 'split':

            
            X_future = _client.scatter(X)
            y_future = _client.scatter(y)
            X_val_future = _client.scatter(X_val)
            y_val_future = _client.scatter(y_val)


            best_pareto_front_idx = list(self.pareto_front.index)
            best_pareto_front = self.pareto_front.loc[best_pareto_front_idx]['Instance']
            val_objective_function_list = [lambda ind, X, y, X_val, y_val, scorers= self._scorers, other_objective_functions=self.other_objective_functions, **kwargs: val_objective_function_generator(
                ind,
                X,y,
                X_val, y_val, 
                scorers= scorers, other_objective_functions=other_objective_functions,
                **kwargs,
                )]
            
            val_scores = tpot2.objectives.parallel_eval_objective_list(
                best_pareto_front,
                val_objective_function_list, n_jobs=self.n_jobs, verbose=self.verbose, timeout=self.max_eval_time_seconds,n_expected_columns=len(self.objective_names),client=_client, X=X_future, y=y_future, X_val=X_val_future, y_val=y_val_future)

            val_objective_names = ['validation_'+name for name in self.objective_names]
            self.objective_names_for_selection = val_objective_names
            self.evaluated_individuals.loc[best_pareto_front_idx,val_objective_names] = val_scores
        else:
            self.objective_names_for_selection = self.objective_names

        val_scores = self.evaluated_individuals[~self.evaluated_individuals[self.objective_names_for_selection].isin(["TIMEOUT","INVALID"]).any(axis=1)][self.objective_names_for_selection].astype(float)                                     
        weighted_scores = val_scores*self.objective_function_weights
        
        if self.bigger_is_better:
            best_idx = weighted_scores[self.objective_names_for_selection[0]].idxmax()
        else:
            best_idx = weighted_scores[self.objective_names_for_selection[0]].idxmin()
        
        best_individual = self.evaluated_individuals.loc[best_idx]['Individual']
        self.selected_best_score =  self.evaluated_individuals.loc[best_idx]
        

        best_individual_pipeline = best_individual.export_pipeline(memory=self.memory, cross_val_predict_cv=self.cross_val_predict_cv, subset_column=self.subset_column)

        if self.preprocessing:
            self.fitted_pipeline_ = sklearn.pipeline.make_pipeline(sklearn.base.clone(self._preprocessing_pipeline), best_individual_pipeline )
        else:
            self.fitted_pipeline_ = best_individual_pipeline 
        
        self.fitted_pipeline_.fit(X_original,y) #TODO use y_original as well?
        if self.verbose >= 3:
            best_individual.plot()

        if self.client is None: #no client was passed in
            #close cluster and client
            _client.close()
            cluster.close()

        return self
        
    def _estimator_has(attr):
        '''Check if we can delegate a method to the underlying estimator.
        First, we check the first fitted final estimator if available, otherwise we
        check the unfitted final estimator.
        '''
        return  lambda self: (self.fitted_pipeline_ is not None and
            hasattr(self.fitted_pipeline_, attr)
        )



    


    @available_if(_estimator_has('predict'))
    def predict(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        return self.fitted_pipeline_.predict(X,**predict_params)
    
    @available_if(_estimator_has('predict_proba'))
    def predict_proba(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        return self.fitted_pipeline_.predict_proba(X,**predict_params)
    
    @available_if(_estimator_has('decision_function'))
    def decision_function(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        return self.fitted_pipeline_.decision_function(X,**predict_params)
    
    @available_if(_estimator_has('transform'))
    def transform(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        return self.fitted_pipeline_.transform(X,**predict_params)

    @property
    def classes_(self):
        """The classes labels. Only exist if the last step is a classifier."""
        return self.fitted_pipeline_.classes_


    def make_evaluated_individuals(self):
        #check if _evolver_instance exists
        if self.evaluated_individuals is None:
            self.evaluated_individuals  =  self._evolver_instance.population.evaluated_individuals.copy()
            objects = list(self.evaluated_individuals.index)
            object_to_int = dict(zip(objects, range(len(objects))))
            self.evaluated_individuals = self.evaluated_individuals.set_index(self.evaluated_individuals.index.map(object_to_int))
            self.evaluated_individuals['Parents'] = self.evaluated_individuals['Parents'].apply(lambda row: _convert_parents_tuples_to_integers(row, object_to_int))

            self.evaluated_individuals["Instance"] = self.evaluated_individuals["Individual"].apply(lambda ind: _apply_make_pipeline(ind, preprocessing_pipeline=self._preprocessing_pipeline))

        return self.evaluated_individuals
        
    @property
    def pareto_front(self):
        #check if _evolver_instance exists
        if self.evaluated_individuals is None:
            return None
        else:
            if "Pareto_Front" not in self.evaluated_individuals:
                return self.evaluated_individuals
            else:
                return self.evaluated_individuals[self.evaluated_individuals["Pareto_Front"]==0]


def get_pareto_front(df, column_names, weights, invalid_values=["TIMEOUT","INVALID"], inplace=True, top=None):
    dftmp = df[~df[column_names].isin(invalid_values).any(axis=1)]

    if "Budget" in dftmp.columns:
        #get rows with the max budget
        dftmp = dftmp[dftmp["Budget"]==dftmp["Budget"].max()]


    indeces = dftmp[~dftmp[column_names].isna().any(axis=1)].index.values
    weighted_scores = df.loc[indeces][column_names].to_numpy()  * weights

    pareto_fronts = tpot2.parent_selectors.nondominated_sorting(weighted_scores)

    if not inplace:
        df = pd.DataFrame(index=df.index,columns=["Pareto_Front"], data=[])
    
    df["Pareto_Front"] = np.nan

    for i, front in enumerate(pareto_fronts):
        for index in front:
            df.loc[indeces[index], "Pareto_Front"] = i

    return df


def _convert_parents_tuples_to_integers(row, object_to_int):
    if type(row) == list or type(row) == np.ndarray or type(row) == tuple:
        return tuple(object_to_int[obj] for obj in row)
    else:
        return np.nan

def _apply_make_pipeline(graphindividual, preprocessing_pipeline=None):
    try: 
        if preprocessing_pipeline is None:
            return graphindividual.export_pipeline()
        else:
            return sklearn.pipeline.make_pipeline(sklearn.base.clone(preprocessing_pipeline), graphindividual.export_pipeline())
    except:
        return None

def get_configuration_dictionary(options, n_samples, n_features, classification, subsets=None, feature_names=None):
    if options is None:
        return options

    if isinstance(options, dict):
        return recursive_with_defaults(options, n_samples, n_features, classification, subsets=subsets, feature_names=feature_names)
    
    if not isinstance(options, list):
        options = [options]

    config_dict = {}

    for option in options:

        if option == "selectors":
            config_dict.update(tpot2.config.make_selector_config_dictionary(classification))

        elif option == "classifiers":
            config_dict.update(tpot2.config.make_classifier_config_dictionary(n_samples=n_samples))

        elif option == "regressors":
            config_dict.update(tpot2.config.make_regressor_config_dictionary(n_samples=n_samples))

        elif option == "transformers":
            config_dict.update(tpot2.config.make_transformer_config_dictionary(n_features=n_features))
        
        elif option == "arithmetic_transformer":
            config_dict.update(tpot2.config.make_arithmetic_transformer_config_dictionary())

        elif option == "feature_set_selector":
            config_dict.update(tpot2.config.make_FSS_config_dictionary(subsets, n_features, feature_names=feature_names))

        elif option == "skrebate":
            config_dict.update(tpot2.config.make_skrebate_config_dictionary(n_features=n_features))
        
        elif option == "MDR":
            config_dict.update(tpot2.config.make_MDR_config_dictionary())
        
        elif option == "ContinuousMDR":
            config_dict.update(tpot2.config.make_ContinuousMDR_config_dictionary())

        elif option == "FeatureEncodingFrequencySelector":
            config_dict.update(tpot2.config.make_FeatureEncodingFrequencySelector_config_dictionary())

        elif option == "genetic encoders":
            config_dict.update(tpot2.config.make_genetic_encoders_config_dictionary())

        elif option == "passthrough":
            config_dict.update(tpot2.config.make_passthrough_config_dictionary())
        

        else:
            config_dict.update(recursive_with_defaults(option, n_samples, n_features, classification, subsets=subsets, feature_names=feature_names))

    if len(config_dict) == 0:
        raise ValueError("No valid configuration options were provided. Please check the options you provided and try again.")

    return config_dict

def recursive_with_defaults(config_dict, n_samples, n_features, classification, subsets=None, feature_names=None):
    
    for key in 'leaf_config_dict', 'root_config_dict', 'inner_config_dict', 'Recursive':
        if key in config_dict:
            value = config_dict[key]
            if key=="Resursive":
                config_dict[key] = recursive_with_defaults(value,n_samples, n_features, classification, subsets=None, feature_names=None)
            else:
                config_dict[key] = get_configuration_dictionary(value, n_samples, n_features, classification, subsets, feature_names)
        
    return config_dict



def objective_function_generator(pipeline, x,y, scorers, cv, other_objective_functions, memory=None, cross_val_predict_cv=None, subset_column=None, step=None, budget=None, generation=1,is_classification=True):
    pipeline = pipeline.export_pipeline(memory=memory, cross_val_predict_cv=cross_val_predict_cv, subset_column=subset_column)
    if budget is not None and budget < 1:
        if is_classification:
            x,y = sklearn.utils.resample(x,y, stratify=y, n_samples=int(budget*len(x)), replace=False, random_state=1)
        else:
            x,y = sklearn.utils.resample(x,y, n_samples=int(budget*len(x)), replace=False, random_state=1)

    if len(scorers) > 0:
        cv_obj_scores = tpot2.estimator_objective_functions.cross_val_score_objective(sklearn.base.clone(pipeline),x,y,scorers=scorers, cv=cv , fold=step)
    else:
        cv_obj_scores = []
    
    if other_objective_functions is not None and len(other_objective_functions) >0:
        other_scores = [obj(sklearn.base.clone(pipeline)) for obj in other_objective_functions]
        #flatten
        other_scores = np.array(other_scores).flatten().tolist()
    else:
        other_scores = []
        
    return np.concatenate([cv_obj_scores,other_scores])


def val_objective_function_generator(pipeline, X_train, y_train, X_test, y_test, scorers, other_objective_functions, memory=None, cross_val_predict_cv=None, subset_column=None, ):
    #subsample the data
    pipeline = pipeline.export_pipeline(memory=memory, cross_val_predict_cv=cross_val_predict_cv, subset_column=subset_column)
    fitted_pipeline = sklearn.base.clone(pipeline)
    fitted_pipeline.fit(X_train, y_train)

    this_fold_scores = [sklearn.metrics.get_scorer(scorer)(fitted_pipeline, X_test, y_test) for scorer in scorers] 
    
    other_scores = []
    #TODO use same exported pipeline as for each objective
    if other_objective_functions is not None and len(other_objective_functions) >0:
        other_scores = [obj(sklearn.base.clone(pipeline)) for obj in other_objective_functions]
    
    return np.concatenate([this_fold_scores,other_scores])


import numpy as np
import random


def random_selector(scores,  k, n_parents=1,): 
    chosen = random.choices(list(range(0,len(scores))), k=k*n_parents)
    return np.reshape(chosen, (k, n_parents))

def max_weighted_average_selector(scores,k, n_parents=1,): 
    ave_scores = [np.nanmean(s ) for s in scores ] #TODO make this more efficient
    chosen = np.argsort(ave_scores)[::-1][0:k] #TODO check this behavior with nans
    return np.reshape(chosen, (k, n_parents))

#based on deap
def TournamentSelection_Dominated(scores, k, n_parents=2):
    """Select the best individual among *tournsize* randomly chosen
    individuals, *k* times. The returned list contains the indices of the chosen *individuals*.
    :param scores: The score matrix, where rows the individulas and the columns are the corresponds to scores on different objectives.
    :param k: The number of individuals to select.
    :param tournsize: The number of individuals participating in each tournament.
    :param score_index: The number of individuals participating in each tournament.
    :returns: A list of indices of selected individuals.
    This function uses the :func:`~random.choice` function from the python base
    :mod:`random` module.
    """
    pareto_fronts = nondominated_sorting(scores)

    # chosen = list(itertools.chain.from_iterable(fronts))
    # if len(chosen) >= k:
    #     return chosen[0:k]

    crowding_dict = {}
    chosen = []
    current_front_number = 0
    while current_front_number < len(pareto_fronts):

        current_front = np.array(list(pareto_fronts[current_front_number]))
        front_scores = [scores[i] for i in current_front]
        crowding_distances = crowding_distance(front_scores)
        for i, crowding in zip(current_front,crowding_distances):
            crowding_dict[i] = crowding

        current_front_number += 1


    chosen = []
    for i in range(k*n_parents):
        asp1 = random.randrange(len(scores))
        asp2 = random.randrange(len(scores))

        if dominates(scores[asp1], scores[asp2]):
            chosen.append(asp1)
        elif dominates(scores[asp2], scores[asp1]):
            chosen.append(asp2)
        
        elif crowding_dict[asp1] > crowding_dict[asp2]:
            chosen.append(asp1)
        elif crowding_dict[asp1] < crowding_dict[asp2]:
            chosen.append(asp2)

        else:
            chosen.append(random.choice([asp1,asp2]))
    
    return np.reshape(chosen, (k, n_parents))


def TournamentSelection(scores, k, n_parents=1, tournament_size=2, score_index=0):
    """Select the best individual among *tournsize* randomly chosen
    individuals, *k* times. The returned list contains the indices of the chosen *individuals*.
    :param scores: The score matrix, where rows the individulas and the columns are the corresponds to scores on different objectives.
    :param k: The number of individuals to select.
    :param tournsize: The number of individuals participating in each tournament.
    :param score_index: The number of individuals participating in each tournament.
    :returns: A list of indices of selected individuals.
    This function uses the :func:`~random.choice` function from the python base
    :mod:`random` module.
    """

    if isinstance(score_index,int):
        key=lambda x:x[1][score_index]
    elif score_index == "average":
        key=lambda x:np.mean(x[1])

    chosen = []
    for i in range(k*n_parents):
        aspirants_idx =[random.randrange(len(scores)) for i in range(tournament_size)]
        aspirants  = list(zip(aspirants_idx, scores[aspirants_idx])) # Zip indices and elements together
        chosen.append(max(aspirants, key=key)[0]) # Retrun the index of the maximum element
    
    return np.reshape(chosen, (k, n_parents))


def LexicaseSelection(scores, k, n_parents=1,):
    """Select the best individual according to Lexicase Selection, *k* times. 
    The returned list contains the indices of the chosen *individuals*.
    :param scores: The score matrix, where rows the individulas and the columns are the corresponds to scores on different objectives.
    :returns: A list of indices of selected individuals.
    This function uses the :func:`~random.choice` function from the python base
    :mod:`random` module.
    """
    chosen =[]
    for i in range(k*n_parents):
        candidates = list(range(len(scores)))
        cases = list(range(len(scores[0])))
        random.shuffle(cases)
        
        while len(cases) > 0 and len(candidates) > 1:
            best_val_for_case = max(scores[candidates,cases[0]])
            candidates = [x for x in candidates if scores[x, cases[0]] == best_val_for_case]
            cases.pop(0)
        chosen.append(random.choice(candidates))

    return np.reshape(chosen, (k, n_parents))


# Deb, Pratab, Agarwal, and Meyarivan, “A fast elitist non-dominated sorting genetic algorithm for multi-objective optimization: NSGA-II”, 2002.
# chatgpt

def nondominated_sorting(matrix):
    """
    Returns the indexes of the matrix 
    bigger is better
    """
    # Initialize the front list and the rank list

    # Initialize the current front
    fronts = {0:set()}

    # Initialize the list of dominated points
    dominated = [set() for _ in range(len(matrix))] #si the set of solutions which solution i dominates

    # Initialize the list of points that dominate the current point
    dominating = [0 for _ in range(len(matrix))] #ni the number of solutions that denominate solution i

    
    # Iterate over all points
    for p, p_scores in enumerate(matrix):
        # Iterate over all other points
        for q, q_scores in enumerate(matrix):
            # If the current point dominates the other point, increment the count of points dominated by the current point
            if dominates(p_scores, q_scores):
                dominated[p].add(q)
            # If the current point is dominated by the other point, add it to the list of dominated points
            elif dominates(q_scores, p_scores):
                dominating[p] += 1
        
        if dominating[p] == 0:
            fronts[0].add(p)

    i=0

    # Iterate until all points have been added to a front
    while len(fronts[i]) > 0:
        H = set()
        for p in fronts[i]:
            for q in dominated[p]:
                dominating[q] -= 1
                if dominating[q] == 0:
                    H.add(q)

        i += 1
        fronts[i] = H


    return [fronts[j] for j in range(i)]


def dominates(list1, list2):
    """
    returns true is all values in list1 are not strictly worse than list2 AND at least one item in list1 is better than list2
    """
    return all(list1[i] >= list2[i] for i in range(len(list1))) and any(list1[i] > list2[i] for i in range(len(list1)))

#adapted from deap + gtp
#bigger is better
def crowding_distance(matrix):
    matrix = np.array(matrix)
    # Initialize the crowding distance for each point to zero
    crowding_distances = [0 for _ in range(len(matrix))]
    
    # Iterate over each objective
    for objective_i in range(matrix.shape[1]):
        # Sort the points according to the current objective
        sorted_i = matrix[:, objective_i].argsort()
        
        # Set the crowding distance of the first and last points to infinity
        crowding_distances[sorted_i[0]] = float("inf")
        crowding_distances[sorted_i[-1]] = float("inf")
        
        if matrix[sorted_i[0]][objective_i] == matrix[sorted_i[-1]][objective_i]: # https://github.com/DEAP/deap/blob/f2a570567fa3dce156d7cfb0c50bc72f133258a1/deap/tools/emo.py#L135
            continue

        norm = matrix.shape[1] * float(matrix[sorted_i[0]][objective_i] - matrix[sorted_i[-1]][objective_i])
        for prev, cur, following in zip(sorted_i[:-2], sorted_i[1:-1], sorted_i[2:]):
            crowding_distances[cur] += (matrix[following][objective_i] - matrix[prev][objective_i]) / norm


    return crowding_distances




def survival_select_NSGA2(scores, k,):

    pareto_fronts = nondominated_sorting(scores)

    # chosen = list(itertools.chain.from_iterable(fronts))
    # if len(chosen) >= k:
    #     return chosen[0:k]

    chosen = []
    current_front_number = 0
    while len(chosen) < k and current_front_number < len(pareto_fronts):

        current_front = np.array(list(pareto_fronts[current_front_number]))
        front_scores = [scores[i] for i in current_front]
        crowding_distances = crowding_distance(front_scores)

        sorted_indeces = current_front[np.argsort(crowding_distances)[::-1]]

        chosen.extend(sorted_indeces[0:(k-len(chosen))])

        current_front_number += 1
    
    return chosen





from .feature_set_selector import FeatureSetSelector
from .zero_count import ZeroCount
from .one_hot_encoder import OneHotEncoder
from .cat_one_hot_encoder import CatOneHotEncoder
from .arithmetictransformer import ArithmeticTransformer
from .passthrough import Passthrough
from .numeric_impute import NumericImpute
from .cat_impute import CatImpute


import numpy as np
import copy
import copy
import typing
import tpot2
from tpot2.individual import BaseIndividual
from traitlets import Bool
import collections
import pandas as pd
from joblib import Parallel, delayed
import copy
import pickle
import dask

def mutate(individual):
    if isinstance(individual, collections.abc.Iterable):
        for ind in individual:
            ind.mutate()
    else:
        individual.mutate()
    return individual

def crossover(parents):
    parents[0].crossover(parents[1])
    return parents[0]

def mutate_and_crossover(parents):
    parents[0].crossover(parents[1])
    parents[0].mutate()
    parents[1].mutate()
    return parents

def crossover_and_mutate(parents):
    for p in parents:
        p.mutate()
    parents[0].crossover(parents[1])
    return parents[0]


built_in_var_ops_dict = {"mutate":mutate, 
                        "crossover":crossover, 
                        "mutate_then_crossover":mutate_and_crossover, 
                        "crossover_then_mutate":crossover_and_mutate}



    
class Population():
    '''
    Primary usage is to keep track of evaluated individuals
    
    Parameters
    ----------
    initial_population : {list of BaseIndividuals}, default=None
        Initial population to start with. If None, start with an empty population.
    use_unique_id : {Bool}, default=True
        If True, individuals are treated as unique if they have the same unique_id().
        If False, all new individuals are treated as unique.
    callback : {function}, default=None
        NOT YET IMPLEMENTED
        A function to call after each generation. The function should take a Population object as its only argument.
    
    Attributes
    ----------
    population : {list of BaseIndividuals}
        The current population of individuals. Contains the live instances of BaseIndividuals.
    evaluated_individuals : {dict}
        A dictionary of dictionaries. The keys are the unique_id() or self of each BaseIndividual.
        Can be thought of as a table with the unique_id() as the row index and the inner dictionary keys as the columns.
    '''
    def __init__(   self,
                    column_names: typing.List[str] = None,
                    n_jobs: int = 1,
                    callback=None,
                    ) -> None:

        if column_names is not None:
            
            column_names = column_names+["Parents", "Variation_Function"]
        else:
            column_names = ["Parents", "Variation_Function"]
        self.evaluated_individuals = pd.DataFrame(columns=column_names)
        self.evaluated_individuals["Parents"] = self.evaluated_individuals["Parents"].astype('object')
        self.use_unique_id = True #Todo clean this up. perhaps pull unique_id() out of baseestimator and have it be supplied as a function
        self.n_jobs = n_jobs
        self.callback=callback
        self.population = []



    #remove individuals that either do not have a column_name value or a nan in that value
    #TODO take into account when the value is not a list/tuple?
    #TODO make invalid a global variable?
    def remove_invalid_from_population(self, column_names, invalid_value = "INVALID"):
        '''
        Remove individuals from the live population if either do not have a value in the column_name column or if the value contains np.nan.
        
        Parameters
        ----------
        column_name : {str}
            The name of the column to check for np.nan values.
        
        Returns
        -------
        None
        '''
        if isinstance(column_names, str): #TODO check this
            column_names = [column_names]
        new_pop = []
        is_valid = lambda ind: ind.unique_id() not in self.evaluated_individuals.index or invalid_value not in self.evaluated_individuals.loc[ind.unique_id(),column_names].to_list()
        self.population = [ind for ind in self.population if is_valid(ind)]

        

    # takes the list of individuals and adds it to the live population list. 
    # if keep_repeats is False, repeated individuals are not added to the population
    # returns a list of individuals added to the live population  
    #TODO make keep repeats allow for previously evaluated individuals,
    #but make sure that the live population only includes one of each, no repeats
    def add_to_population(self, individuals: typing.List[BaseIndividual], keep_repeats=False, mutate_until_unique=True):
        '''
        Add individuals to the live population. Add individuals to the evaluated_individuals if they are not already there.
        
        Parameters:
        -----------
        individuals : {list of BaseIndividuals}
            The individuals to add to the live population.
        keep_repeats : {bool}, default=False
            If True, allow the population to have repeated individuals.
            If False, only add individuals that have not yet been added to geneology.
        '''
        if not isinstance(individuals, collections.abc.Iterable):
            individuals = [individuals]

        new_individuals = []
        #TODO check for proper inputs
        for individual in individuals:
            key = individual.unique_id()

            if key not in self.evaluated_individuals.index: #If its new, we always add it
                self.evaluated_individuals.loc[key] = np.nan
                self.evaluated_individuals.loc[key,"Individual"] = copy.deepcopy(individual)
                self.population.append(individual)
                new_individuals.append(individual)

            else:#If its old
                if keep_repeats: #If we want to keep repeats, we add it
                    self.population.append(individual)
                    new_individuals.append(individual)
                elif mutate_until_unique: #If its old and we don't want repeats, we can optionally mutate it until it is unique
                    for _ in range(20):
                        individual = copy.deepcopy(individual)
                        individual.mutate()
                        key = individual.unique_id()
                        if key not in self.evaluated_individuals.index:
                            self.evaluated_individuals.loc[key] = np.nan
                            self.evaluated_individuals.loc[key,"Individual"] = copy.deepcopy(individual)
                            self.population.append(individual)
                            new_individuals.append(individual)
                            break
                    
        return new_individuals


    def update_column(self, individual, column_names, data):
        '''
        Update the column_name column in the evaluated_individuals with the data.
        If the data is a list, it must be the same length as the evaluated_individuals.
        If the data is a single value, it will be applied to all individuals in the evaluated_individuals.
        '''
        if isinstance(individual, collections.abc.Iterable):
            if self.use_unique_id:
                key = [ind.unique_id() for ind in individual]
            else:
                key = individual
        else:
            if self.use_unique_id:
                key = individual.unique_id()
            else:
                key = individual

        self.evaluated_individuals.loc[key,column_names] = data

    
    def get_column(self, individual, column_names=None, to_numpy=True):
        '''
        Update the column_name column in the evaluated_individuals with the data.
        If the data is a list, it must be the same length as the evaluated_individuals.
        If the data is a single value, it will be applied to all individuals in the evaluated_individuals.
        '''
        if isinstance(individual, collections.abc.Iterable):
            if self.use_unique_id:
                key = [ind.unique_id() for ind in individual]
            else:
                key = individual
        else:
            if self.use_unique_id:
                key = individual.unique_id()
            else:
                key = individual

        if column_names is not None:
            slice = self.evaluated_individuals.loc[key,column_names]
        else:
            slice = self.evaluated_individuals.loc[key]
        if to_numpy:
            slice.reset_index(drop=True, inplace=True)
            return slice.to_numpy()
        else:
            return slice


    #returns the individuals without a 'column' as a key in geneology
    #TODO make sure not to get repeats in this list even if repeats are in the "live" population
    def get_unevaluated_individuals(self, column_names, individual_list=None):
        if individual_list is None:
            individual_list = self.population
        
        if self.use_unique_id:
            unevaluated_filter = lambda individual: individual.unique_id() not in self.evaluated_individuals.index or any(self.evaluated_individuals.loc[individual.unique_id(), column_names].isna())
        else:
            unevaluated_filter = lambda individual: individual not in self.evaluated_individuals.index or any(self.evaluated_individuals.loc[individual.unique_id(), column_names].isna())
        
        return [individual for individual in individual_list if unevaluated_filter(individual)]    

    # def get_valid_evaluated_individuals_df(self, column_names_to_check, invalid_values=["TIMEOUT","INVALID"]):
    #     '''
    #     Returns a dataframe of the evaluated individuals that do no have invalid_values in column_names_to_check.
    #     '''
    #     return self.evaluated_individuals[~self.evaluated_individuals[column_names_to_check].isin(invalid_values).any(axis=1)]

    #the live population empied and is set to new_population
    def set_population(self,  new_population, keep_repeats=True):
        '''
        sets population to new population
        for selection?
        '''
        self.population = []
        self.add_to_population(new_population, keep_repeats=keep_repeats)

    #TODO should we just generate one offspring per crossover? 
    def create_offspring(self, parents_list, var_op_list, add_to_population=True, keep_repeats=False, mutate_until_unique=True, n_jobs=1):
        '''
        parents_list: a list of lists of parents. 
        var_op_list: a list of var_ops to apply to each list of parents. Should be the same length as parents_list.

        for example:
        parents_list = [[parent1, parent2], [parent3]]
        var_op_list = ["crossover", "mutate"]

        This will apply crossover to parent1 and parent2 and mutate to parent3.

        Creates offspring from parents using the var_op_list.
        If string, will use a built in method 
            - "crossover" : crossover
            - "mutate" : mutate
            - "mutate_and_crossover" : mutate_and_crossover
            - "cross_and_mutate" : cross_and_mutate
        '''
        new_offspring = []
        all_offspring = parallel_create_offspring(parents_list, var_op_list, n_jobs=n_jobs)

        for parents, offspring, var_op in zip(parents_list, all_offspring, var_op_list):
            
            # if var_op in built_in_var_ops_dict:
            #     var_op = built_in_var_ops_dict[var_op]

            # offspring = copy.deepcopy(parents)
            # offspring = var_op(offspring)
            # if isinstance(offspring, collections.abc.Iterable):
            #     offspring = offspring[0] 

            if add_to_population:
                added = self.add_to_population(offspring, keep_repeats=keep_repeats, mutate_until_unique=mutate_until_unique)
                if len(added) > 0:
                    for new_child in added:
                        parent_keys = [parent.unique_id() for parent in parents]
                        if not pd.api.types.is_object_dtype(self.evaluated_individuals["Parents"]): #TODO Is there a cleaner way of doing this? Not required for some python environments?
                            self.evaluated_individuals["Parents"] = self.evaluated_individuals["Parents"].astype('object')
                        self.evaluated_individuals.at[new_child.unique_id(),"Parents"] = tuple(parent_keys)
                        
                        #if var_op is a function
                        if hasattr(var_op, '__call__'):
                            self.evaluated_individuals.at[new_child.unique_id(),"Variation_Function"] = var_op.__name__
                        else:
                            self.evaluated_individuals.at[new_child.unique_id(),"Variation_Function"] = var_op
                        
                        
                        new_offspring.append(new_child)

            else:
                new_offspring.append(offspring)
                            
            
        return new_offspring

   

def get_id(individual):
    return individual.unique_id()

def parallel_create_offspring(parents_list, var_op_list, n_jobs=1):
    delayed_offspring = []
    for parents, var_op in zip(parents_list,var_op_list):
        #TODO put this loop in population class
        if var_op in built_in_var_ops_dict:
            var_op = built_in_var_ops_dict[var_op]
        delayed_offspring.append(dask.delayed(copy_and_change)(parents, var_op))

    offspring = dask.compute(*delayed_offspring,
                             num_workers=n_jobs, threads_per_worker=1)
    return offspring

def copy_and_change(parents, var_op):
    offspring = copy.deepcopy(parents)
    offspring = var_op(offspring)
    if isinstance(offspring, collections.abc.Iterable):
        offspring = offspring[0]
    return offspring

def parallel_get_id(n_jobs, individual_list):
    id_list = Parallel(n_jobs=n_jobs)(delayed(get_id)(ind)  for ind in individual_list)
    return id_list



# -*- coding: utf-8 -*-

"""Copyright (c) 2015 The auto-sklearn developers. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of the auto-sklearn Developers  nor the names of
     its contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""
#TODO support np arrays

import numpy as np
from scipy import sparse

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.preprocessing import OneHotEncoder
import sklearn
import sklearn.impute

import pandas as pd
from pandas.api.types import is_numeric_dtype



def auto_select_categorical_features(X):

    if not isinstance(X, pd.DataFrame):
        return []
    
    feature_mask = []
    for column in X.columns:
        feature_mask.append(not is_numeric_dtype(X[column]))

    return feature_mask


def _X_selected(X, selected):
    """Split X into selected features and other features"""
    X_sel = X[X.columns[selected]]
    X_not_sel = X.drop(X.columns[selected], axis=1)
    return X_sel, X_not_sel



class CatImpute(BaseEstimator, TransformerMixin):


    def __init__(self, categorical_features='auto'):
        self.categorical_features = categorical_features


    def fit(self, X, y=None):
        """Fit OneHotEncoder to X, then transform X.

        Equivalent to self.fit(X).transform(X), but more convenient and more
        efficient. See fit for the parameters, transform for the return value.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Dense array or sparse matrix.
        y: array-like {n_samples,} (Optional, ignored)
            Feature labels
        """
        
        if self.categorical_features == "auto":
            self.categorical_features_ = auto_select_categorical_features(X)

        if sum(self.categorical_features_) == 0:
            return self

        #TODO make this more consistent with sklearn baseimputer/baseencoder
        if isinstance(X, pd.DataFrame):
            for col in X.columns:
                # check if the column name is not a string
                if not isinstance(col, str):
                    # if it's not a string, rename the column with "X" prefix
                    X.rename(columns={col: f"X{col}"}, inplace=True)

        self.enc = sklearn.impute.SimpleImputer(strategy='most_frequent')
        if isinstance(X, pd.DataFrame):
            self.enc.set_output(transform="pandas")

        if sum(self.categorical_features_) == X.shape[1]:
            
            X_sel = self.enc.fit(X)
        else:
            X_sel, X_not_sel = _X_selected(X, self.categorical_features_)
            X_sel = self.enc.fit(X_sel)
        
        return self
  
    def transform(self, X):
        """Transform X using one-hot encoding.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Dense array or sparse matrix.

        Returns
        -------
        X_out : sparse matrix if sparse=True else a 2-d array, dtype=int
            Transformed input.
        """

    
        if sum(self.categorical_features_) == 0:
            return X

        #TODO make this more consistent with sklearn baseimputer/baseencoder
        if isinstance(X, pd.DataFrame):
            for col in X.columns:
                # check if the column name is not a string
                if not isinstance(col, str):
                    # if it's not a string, rename the column with "X" prefix
                    X.rename(columns={col: f"X{col}"}, inplace=True)

        if sum(self.categorical_features_) == X.shape[1]:
            return self.enc.transform(X)
        else:

            X_sel, X_not_sel= _X_selected(X, self.categorical_features_)
            X_sel = self.enc.transform(X_sel)
            
            #If X is dataframe
            if isinstance(X, pd.DataFrame):
            
                X_sel = pd.DataFrame(X_sel, columns=self.enc.get_feature_names_out())
                return pd.concat([X_not_sel.reset_index(drop=True), X_sel.reset_index(drop=True)], axis=1)
            else:
                return np.hstack((X_not_sel, X_sel))



from functools import partial
from tpot2.builtin_modules import ZeroCount, OneHotEncoder
import numpy as np


def params_sklearn_preprocessing_Binarizer(trial, name=None):
    return {
        'threshold': trial.suggest_float(f'threshold_{name}', 0.0, 1.0),
    }

def params_sklearn_decomposition_FastICA(trial, name=None, n_features=100):
    return {
        'algorithm': trial.suggest_categorical(f'algorithm_{name}', ['parallel', 'deflation']),
        'whiten':'unit-variance',
    }

def params_sklearn_cluster_FeatureAgglomeration(trial, name=None, n_features=100):
    return {
        'metric': trial.suggest_categorical(f'metric_{name}', ['euclidean', 'l1', 'l2', 'manhattan', 'cosine']),
        'linkage': trial.suggest_categorical(f'linkage_{name}', ['ward', 'complete', 'average']),
        'n_clusters': trial.suggest_int(f'n_clusters_{name}', 2, 4), #TODO perhaps a percentage of n_features
    }



def params_sklearn_preprocessing_Normalizer(trial, name=None):
    return {
        'norm': trial.suggest_categorical(f'norm_{name}', ['l1', 'l2', 'max']),
    }

def params_sklearn_kernel_approximation_Nystroem(trial, name=None, n_features=100):
    return {
        'gamma': trial.suggest_float(f'gamma_{name}', 0.0, 1.0),
        'kernel': trial.suggest_categorical(f'kernel_{name}', ['rbf', 'cosine', 'chi2', 'laplacian', 'polynomial', 'poly', 'linear', 'additive_chi2', 'sigmoid']),
        'n_components': trial.suggest_int(f'n_components_{name}', 1, 11), #TODO perhaps a percentage of n_features
    }

def params_sklearn_decomposition_PCA(trial, name=None, n_features=100):
    # keep the number of components required to explain 'variance_explained' of the variance
    variance_explained = 1 - trial.suggest_float(f'n_components_{name}', 0.001, 0.5, log=True) #values closer to 1 are more likely
    
    return {
        'n_components': variance_explained,
    }


def params_sklearn_kernel_approximation_RBFSampler(trial, name=None, n_features=100):
    return {
        'gamma': trial.suggest_float(f'gamma_{name}', 0.0, 1.0),
    }




def params_tpot_builtins_ZeroCount(trial, name=None):

    return {}
    

def params_tpot_builtins_OneHotEncoder(trial, name=None):

    return {}


    
from sklearn.preprocessing import Binarizer
from sklearn.decomposition import FastICA
from sklearn.cluster import FeatureAgglomeration
from sklearn.preprocessing import MaxAbsScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import Normalizer
from sklearn.kernel_approximation import Nystroem
from sklearn.decomposition import PCA
from sklearn.preprocessing import PolynomialFeatures
from sklearn.kernel_approximation import RBFSampler
from sklearn.preprocessing import RobustScaler
from sklearn.preprocessing import StandardScaler

def make_transformer_config_dictionary(n_features=10):
    #n_features = min(n_features,100) #TODO optimize this
    return {
                Binarizer: params_sklearn_preprocessing_Binarizer,
                FastICA: partial(params_sklearn_decomposition_FastICA,n_features=n_features),
                FeatureAgglomeration: partial(params_sklearn_cluster_FeatureAgglomeration,n_features=n_features),
                MaxAbsScaler: {},
                MinMaxScaler: {},
                Normalizer: params_sklearn_preprocessing_Normalizer,
                Nystroem: partial(params_sklearn_kernel_approximation_Nystroem,n_features=n_features),
                PCA: partial(params_sklearn_decomposition_PCA,n_features=n_features),
                PolynomialFeatures: {
                                        'degree': 2,
                                        'include_bias': False,
                                        'interaction_only': False,
                                    },
                RBFSampler: partial(params_sklearn_kernel_approximation_RBFSampler,n_features=n_features),
                RobustScaler: {},
                StandardScaler: {},
                ZeroCount: params_tpot_builtins_ZeroCount,
                OneHotEncoder: params_tpot_builtins_OneHotEncoder,
            }





import numpy as np
from scipy import sparse

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.preprocessing import OneHotEncoder
import sklearn

import pandas as pd
from pandas.api.types import is_numeric_dtype



def auto_select_categorical_features(X):

    if not isinstance(X, pd.DataFrame):
        return []
    
    feature_mask = []
    for column in X.columns:
        feature_mask.append(not is_numeric_dtype(X[column]))

    return feature_mask


def _X_selected(X, selected):
    """Split X into selected features and other features"""
    X_sel = X[X.columns[selected]]
    X_not_sel = X.drop(X.columns[selected], axis=1)
    return X_sel, X_not_sel



class CatOneHotEncoder(BaseEstimator, TransformerMixin):


    def __init__(self, categorical_features='auto'):
        self.categorical_features = categorical_features


    def fit(self, X, y=None):
        """Fit OneHotEncoder to X, then transform X.

        Equivalent to self.fit(X).transform(X), but more convenient and more
        efficient. See fit for the parameters, transform for the return value.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Dense array or sparse matrix.
        y: array-like {n_samples,} (Optional, ignored)
            Feature labels
        """
        
        if self.categorical_features == "auto":
            self.categorical_features_ = auto_select_categorical_features(X)

        if sum(self.categorical_features_) == 0:
            return self
        
        self.enc = sklearn.preprocessing.OneHotEncoder(categories='auto', handle_unknown='ignore',sparse_output=False)

        #TODO make this more consistent with sklearn baseimputer/baseencoder
        if isinstance(X, pd.DataFrame):
            self.enc.set_output(transform="pandas")
            for col in X.columns:
                # check if the column name is not a string
                if not isinstance(col, str):
                    # if it's not a string, rename the column with "X" prefix
                    X.rename(columns={col: f"X{col}"}, inplace=True)


        if sum(self.categorical_features_) == X.shape[1]:
            X_sel = self.enc.fit(X)
        else:
            X_sel, X_not_sel = _X_selected(X, self.categorical_features_)
            X_sel = self.enc.fit(X_sel)
        
        return self
  
    def transform(self, X):
        """Transform X using one-hot encoding.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Dense array or sparse matrix.

        Returns
        -------
        X_out : sparse matrix if sparse=True else a 2-d array, dtype=int
            Transformed input.
        """

    
        if sum(self.categorical_features_) == 0:
            return X
        
        #TODO make this more consistent with sklearn baseimputer/baseencoder
        if isinstance(X, pd.DataFrame):
            for col in X.columns:
                # check if the column name is not a string
                if not isinstance(col, str):
                    # if it's not a string, rename the column with "X" prefix
                    X.rename(columns={col: f"X{col}"}, inplace=True)

        if sum(self.categorical_features_) == X.shape[1]:
            return self.enc.transform(X)
        else:

            X_sel, X_not_sel= _X_selected(X, self.categorical_features_)
            X_sel = self.enc.transform(X_sel)
            
            #If X is dataframe
            if isinstance(X, pd.DataFrame):
            
                X_sel = pd.DataFrame(X_sel, columns=self.enc.get_feature_names_out())
                return pd.concat([X_not_sel.reset_index(drop=True), X_sel.reset_index(drop=True)], axis=1)
            else:
                return np.hstack((X_not_sel, X_sel))




#TODO: are all the imports in the init files done correctly?
#TODO clean up import organization
from ._version import __version__
from .graphsklearn import *
from .individual import *
from .logbook import *
from .population import *
from .base_evolver import *
from .objectives import *
from .evolutionary_algorithms import *
from .tpot_estimator import *
from .builtin_modules import *
from .config import *
from .representations import *
from .parent_selectors import *